#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "util.h"
#include "types.h"
#include "ArchTokens.h"
#include "ReadArch.h"


/******** Function Prototypes ********/

static void Error(IN const char *Token,
				  IN int LineNum);
static void ErrorDupe(IN const char *Token,
				      IN int LineNum);
static const char *const * ReadArgs(IN const char *const *Args,
									OUT t_arch *Arch, 
									IN int LineNum);
static const char *const * ReadInt(IN const char *const *Str, 
								   OUT int *Val,
								   IN int LineNum);
static const char *const * ReadFloat(IN const char *const *Str, 
								     OUT float *Val,
									 IN int LineNum);
static const char *const * ReadChanWidth(IN const char *const *Args,
						  				 OUT t_chan *Chan,
										 IN int LineNum);
static const char *const * ReadToken(IN const char *const *Args,
									 OUT enum e_ArchToken *Token,
									 IN int LineNum);
static const char *const * ReadBaseToken(IN const char *const *Args,
										 OUT enum e_ArchBaseToken *Token,
										 IN int LineNum);
static const char *const * ReadPin(IN const char *const *Args,
								   INOUT struct s_pin **PinList,
								   IN enum e_pin_type Type,
								   IN int LineNum);
static const char *const * ReadFcType(IN const char *const *Args,
									  OUT enum e_Fc_type *Val,
									  IN int LineNum);
static const char *const * ReadSBType(IN const char *const *Args,
									  OUT enum e_switch_block_type *Val,
									  IN int LineNum);
static const char *const * ReadCLBGrid(IN const char *const *Args,
									   OUT struct s_clb_grid *clb_grid,
									   IN int LineNum);
static const char *const * ReadFcOut(IN const char *const *Args,
									   OUT float *Fc_out,
									   OUT boolean *Fc_out_full_flex,
									   IN int LineNum);
static const char *const * ReadSegment(IN const char *const *Args,
									   INOUT struct s_seg **SegList,
									   IN int LineNum);
static const char *const * ReadSwitch(IN const char *const *Args,
									  INOUT struct s_switch **SwitchList,
									  IN int LineNum);
static const char *const * ReadTSubblock(IN const char *const *Args,
										 INOUT struct s_tsubblock **TSubblockList,
										 IN int LineNum);
static const char *const * ReadTag(IN const char *const *Args,
								   IN enum e_ArchToken Token,
								   IN int LineNum);



/******** Subroutines ********/


/* Called for parse errors. Spits out a message and then exits program. */
static void Error(IN const char *Token,
				  IN int LineNum) {
	if (Token) {
		printf(ERRTAG "Unexpected token '%s' on line %d of architecture file\n",
		       Token, LineNum);
	} 
	else {
		printf(ERRTAG "Missing token at end of line %d of architecture file\n",
		       LineNum);
	}
	exit(1);
}


static void ErrorDupe(IN const char *Token,
				      IN int LineNum) {
	printf(ERRTAG "Option '%s' specified more than once on line %d of architecture file\n",
	       Token, LineNum);
	exit(1);
}


/* This is the subroutine that actually loads the arch params.
 * Returns number of tokens consumed so caller can look for extras */
static const char *const * ReadArgs(IN const char *const *Args,
									OUT t_arch *Arch, 
									IN int LineNum) {
	const char *const *PrevArgs;
	enum e_ArchBaseToken Token;

	PrevArgs = Args;
	Args = ReadBaseToken(Args, &Token, LineNum);

	/* Count the number of times each command appears */
	if (Token < AT_BASE_UNKNOWN) {
		++Arch->Count[Token];
	}

	switch (Token) {
		case AT_IO_RAT: return ReadInt(Args, &Arch->IO_Ratio, LineNum);
		case AT_CHAN_WIDTH_X: return ReadChanWidth(Args, &Arch->Chans.chan_x_dist, LineNum);
		case AT_CHAN_WIDTH_Y: return ReadChanWidth(Args, &Arch->Chans.chan_y_dist, LineNum);
		case AT_CHAN_WIDTH_IO: return ReadFloat(Args, &Arch->Chans.chan_width_io, LineNum);
		case AT_OUTPIN: return ReadPin(Args, &Arch->Pins, DRIVER, LineNum);
		case AT_INPIN: return ReadPin(Args, &Arch->Pins, RECEIVER, LineNum);
		case AT_SUBBLOCKS_PER_CLB: return ReadInt(Args, &Arch->N, LineNum);
		case AT_SUBBLOCK_LUT_SIZE: return ReadInt(Args, &Arch->K, LineNum);
		case AT_FC_OUTPUT: return ReadFcOut(Args, &Arch->Fc_Out, &Arch->Fc_out_full_flex, LineNum);
		case AT_FC_INPUT: return ReadFloat(Args, &Arch->Fc_In, LineNum);
		case AT_FC_PAD: return ReadFloat(Args, &Arch->Fc_Pad, LineNum);
		case AT_FC_TYPE: return ReadFcType(Args, &Arch->Fc_Type, LineNum);
		case AT_SWITCH_BLOCK_TYPE: return ReadSBType(Args, &Arch->SBType, LineNum);
		case AT_R_MINW_NMOS: return ReadFloat(Args, &Arch->R_minW_nmos, LineNum);
		case AT_R_MINW_PMOS: return ReadFloat(Args, &Arch->R_minW_pmos, LineNum);
		case AT_FS: return ReadInt(Args, &Arch->Fs, LineNum);
		case AT_C_IPIN_SBLOCK: return ReadFloat(Args, &Arch->C_ipin_cblock, LineNum);
   		case AT_T_IPIN_CBLOCK: return ReadFloat(Args, &Arch->T_ipin_cblock, LineNum);
		case AT_T_IPAD: return ReadFloat(Args, &Arch->T_ipad, LineNum);
		case AT_T_OPAD: return ReadFloat(Args, &Arch->T_opad, LineNum);
		case AT_T_SBLK_OPIN_TO_SBLK_IPIN: return ReadFloat(Args, &Arch->T_sblk_opin_to_sblk_ipin, LineNum);
  		case AT_T_CLB_IPIN_TO_SBLK_IPIN: return ReadFloat(Args, &Arch->T_clb_ipin_to_sblk_ipin, LineNum);
		case AT_T_SBLK_OPIN_TO_CLB_OPIN: return ReadFloat(Args, &Arch->T_sblk_opin_to_clb_opin, LineNum);
		case AT_CLB_GRID: return ReadCLBGrid(Args, &Arch->clb_grid, LineNum);
		case AT_SEGMENT: return ReadSegment(Args, &Arch->Segments, LineNum);
		case AT_SWITCH: return ReadSwitch(Args, &Arch->Switches, LineNum);
		case AT_T_SUBBLOCK: return ReadTSubblock(Args, &Arch->TSubblocks, LineNum);
		case AT_CLB_AREA: return ReadFloat(Args, &Arch->clb_area, LineNum);
		case AT_IPIN_MUX_TRANS_SIZE: return ReadFloat(Args, &Arch->ipin_mux_trans_size, LineNum);
		default: Error(*PrevArgs, LineNum);
	}
	return PrevArgs;
}


void ReadArch(INOUT char *FileName,
			  OUT t_arch *Arch) {

	FILE* ArchFile;
	char **Tokens;
	const char *const *Args;
	int LineNum = 0;

	/* Clear the structure to initialize.  */
	memset(Arch, 0, sizeof(t_arch));

	ArchFile = my_fopen(FileName, "r");

	while (NULL != (Tokens = ReadLineTokens(ArchFile, &LineNum))) {
		Args = ReadArgs((const char *const *)Tokens, Arch, LineNum);

		/* Check for leftover tokens */
		if (*Args) Error(*Args, LineNum);
		
        /* Free token data and token pointers */
		free(*Tokens);
        free(Tokens);
	}

	fclose(ArchFile);
}


static const char *const * ReadTag(IN const char *const *Args,
								   IN enum e_ArchToken Token,
								   IN int LineNum) {
	enum e_ArchToken Cur;
	const char *const *NextArgs;

	NextArgs = ReadToken(Args, &Cur, LineNum);
	if (Cur != Token) Error(*Args, LineNum);
	
	return NextArgs;
}


static const char *const * ReadTSubblock(IN const char *const *Args,
										 INOUT struct s_tsubblock **TSubblockList,
										 IN int LineNum) {
	enum e_ArchToken Token;
	struct s_tsubblock *Cur;
	struct s_tsubblock *Tmp;
	const char *const *PrevArgs;
	bitfield Flag;

	Cur = (struct s_tsubblock*) my_malloc(sizeof(struct s_tsubblock));
	memset(Cur, 0, sizeof(struct s_tsubblock));
	if (*TSubblockList) {
		Tmp = *TSubblockList;
		while (Tmp->Next) Tmp = Tmp->Next;
		Tmp->Next = Cur;
	}
	else {
		*TSubblockList = Cur;
	}

	while (*Args) {
		PrevArgs = Args;
		Args = ReadToken(Args, &Token, LineNum);
		Flag = 0;

		switch (Token) {
			case AT_T_COMB:
				Args = ReadFloat(Args, &Cur->T_comb, LineNum);
				Flag = S_TSUBBLOCK_T_COMB;
				break;

			case AT_T_SEQ_IN:
				Args = ReadFloat(Args, &Cur->T_seq_in, LineNum);
				Flag = S_TSUBBLOCK_T_SEQ_IN;
				break;

			case AT_T_SEQ_OUT:
				Args = ReadFloat(Args, &Cur->T_seq_out, LineNum);
				Flag = S_TSUBBLOCK_T_SEQ_OUT;
				break;

			default: Error(*PrevArgs, LineNum);
		}

		if (Cur->IsSet & Flag) {
			ErrorDupe(*PrevArgs, LineNum);
		}	
		else {
			Cur->IsSet |= Flag;
		}
	}
	return Args;
}


static const char *const * ReadSwitch(IN const char *const *Args,
									  INOUT struct s_switch **SwitchList,
									  IN int LineNum) {
	enum e_ArchToken Token;
	const char *const *PrevArgs;
	const char *const *PrevArgs2;
	struct s_switch *Cur;
	struct s_switch *Tmp;
	bitfield Flag;

	/* New linked list item */
	Cur = (struct s_switch*) my_malloc(sizeof(struct s_switch));
	memset(Cur, 0, sizeof(struct s_switch));
	if (*SwitchList) {
		Tmp = *SwitchList;
		while (Tmp->Next) 
			Tmp = Tmp->Next;
		Tmp->Next = Cur;
	}
	else {
		*SwitchList = Cur;
	}

	/* Switch number comes right after 'segment' */
	Args = ReadInt(Args, &Cur->Index, LineNum);

	/* The rest of the options are in any order */
	while (*Args) {
		PrevArgs = Args;
		Args = ReadToken(Args, &Token, LineNum);
		Flag = 0;
		switch (Token) {
			case AT_BUFFERED:
				PrevArgs2 = Args;
				Args = ReadToken(Args, &Token, LineNum);
				switch (Token) {
					case AT_YES: Cur->Buffered = TRUE; break;
					case AT_NO: Cur->Buffered = FALSE; break;
					default: Error(*PrevArgs2, LineNum);
				}
				Flag = S_SWITCH_BUFFERED;
				break;

			case AT_R:
				Args = ReadFloat(Args, &Cur->R, LineNum);
				Flag = S_SWITCH_R;
				break;

			case AT_CIN:
				Args = ReadFloat(Args, &Cur->Cin, LineNum);
				Flag = S_SWITCH_CIN;
				break;

			case AT_COUT:
				Args = ReadFloat(Args, &Cur->Cout, LineNum);
				Flag = S_SWITCH_COUT;
				break;

			case AT_TDEL:
				Args = ReadFloat(Args, &Cur->Tdel, LineNum);
				Flag = S_SWITCH_TDEL;
				break;

			case AT_BUF_SIZE:
				Args = ReadFloat(Args, &Cur->BufSize, LineNum);
				Flag = S_SWITCH_BUFSIZE;
				break;

			case AT_MUX_TRANS_SIZE:
				Args = ReadFloat(Args, &Cur->MuxTransSize, LineNum);
				Flag = S_SWITCH_MUXTRANSSIZE;
				break;

			default: Error(*PrevArgs, LineNum);
		}

		/* Set flag and error if was already set */
		if (Cur->IsSet & Flag) {
			ErrorDupe(*PrevArgs, LineNum);
		}
		else {
			Cur->IsSet |= Flag;
		}
	}
	return Args;
}


static const char *const * ReadSegment(IN const char *const *Args,
									   INOUT struct s_seg **SegList,
									   IN int LineNum) {
	enum e_ArchToken Token;
	const char *const *PrevArgs;
	const char *const *PrevArgs2;
	struct s_seg *Cur;
	struct s_seg *Tmp;
	bitfield Flag;

	/* Add new item to end of linked list */
	Cur = (struct s_seg*) my_malloc(sizeof(struct s_seg));
	memset(Cur, 0, sizeof(struct s_seg));
	if (*SegList) {
		Tmp = *SegList;
		while (Tmp->Next) 
			Tmp = Tmp->Next;
		Tmp->Next = Cur;
	}
	else {
		*SegList = Cur;
	}

	/* Go through options for the segment */
	while (*Args) {
		PrevArgs = Args;
		Args = ReadToken(Args, &Token, LineNum);
		Flag = 0;
		switch (Token) {
			case AT_FREQUENCY: 
				Args = ReadFloat(Args, &Cur->Freq, LineNum); 
				Flag = S_SEG_FREQUENCY; 
				break;

			case AT_WIRE_SWITCH: 
				Args = ReadInt(Args, &Cur->WireSwitch, LineNum); 
				Flag = S_SEG_WIRE_SWITCH;
				break;

			case AT_OPIN_SWITCH: 
				Args = ReadInt(Args, &Cur->OPinSwitch, LineNum); 
				Flag = S_SEG_OPIN_SWITCH;
				break;

			case AT_FRAC_CB: 
				Args = ReadFloat(Args, &Cur->Frac_cb, LineNum); 
				Flag = S_SEG_FRAC_CB;
				break;

			case AT_FRAC_SB: 
				Args = ReadFloat(Args, &Cur->Frac_sb, LineNum);
				Flag = S_SEG_FRAC_SB;
				break;

			case AT_RMETAL: 
				Args = ReadFloat(Args, &Cur->Rmetal, LineNum); 
				Flag = S_SEG_RMETAL;
				break;

			case AT_CMETAL: 
				Args = ReadFloat(Args, &Cur->Cmetal, LineNum); 
				Flag = S_SEG_CMETAL;
				break;

			case AT_DIRECTIONALITY:
				PrevArgs2 = Args;
				Args = ReadToken(Args, &Token, LineNum);
				switch (Token) {
					case AT_BI_DIRECTIONAL: Cur->Directionality = BI_DIRECTIONAL; break;
					case AT_UNI_DIRECTIONAL: Cur->Directionality = UNI_DIRECTIONAL; break;
					default: Error(*PrevArgs2, LineNum);
				}
				Flag = S_SEG_DIRECTIONALITY;
				break;
				
			case AT_LENGTH: 
				PrevArgs2 = Args;
				Args = ReadToken(Args, &Token, LineNum);
				if (AT_LONGLINE == Token) {
					Cur->IsLong = TRUE;
				}
				else {
					/* Reparse same token as a number instead */
					Args = PrevArgs2;
					Args = ReadInt(Args, &Cur->Len, LineNum); 
					Cur->IsLong = FALSE;
				}
				Flag = S_SEG_LENGTH;
				break;
			
			default: Error(*PrevArgs, LineNum);
		}

		if (Cur->IsSet & Flag) {
			ErrorDupe(*PrevArgs, LineNum);
		}
		else {
			Cur->IsSet |= Flag;
		}
	}

	return Args;	
}


static const char *const * ReadFcOut(IN const char *const *Args,
									   OUT float *Fc_out,
									   OUT boolean *Fc_out_full_flex,
									   IN int LineNum) {
	enum e_ArchToken Token;
	const char* const* PrevArgs = Args;
	Args = ReadToken(Args, &Token, LineNum);
	if (AT_FULL == Token) {
		*Fc_out_full_flex = TRUE;
		*Fc_out = 1.0;
	}
	else {
		/* Reparse same token as a number instead */
		Args = PrevArgs;
		Args = ReadFloat(Args, Fc_out, LineNum); 
		*Fc_out_full_flex = FALSE;
	}
	return Args;
}


static const char *const * ReadCLBGrid(IN const char *const *Args,
									   OUT struct s_clb_grid *clb_grid,
									   IN int LineNum) {
	enum e_ArchToken Token;
	const char* const* PrevArgs = Args;
	Args = ReadToken(Args, &Token, LineNum);
	if (AT_AUTO == Token) {
		Args = ReadFloat(Args, &clb_grid->Aspect, LineNum);
		clb_grid->IsAuto = TRUE; 
	}
	else {
		/* Reload same token as a number instead */	
		Args = PrevArgs;
		Args = ReadInt(Args, &clb_grid->W, LineNum);
		Args = ReadInt(Args, &clb_grid->H, LineNum);
		clb_grid->IsAuto = FALSE;
	}
	return Args;
}


static const char *const * ReadSBType(IN const char *const *Args,
									  OUT enum e_switch_block_type *Val,
									  IN int LineNum) {
	enum e_ArchToken Token;
	const char* const* PrevArgs = Args;
	Args = ReadToken(Args, &Token, LineNum);
	switch(Token) {
		case AT_SUBSET: *Val = SUBSET; break;
		case AT_WILTON: *Val = WILTON; break;
		case AT_UNIVERSAL: *Val = UNIVERSAL; break;
		case AT_FULL: *Val = FULL; break;
		default: Error(*PrevArgs, LineNum); break;
	}
	return Args;
}


static const char *const * ReadFcType(IN const char *const *Args,
									  OUT enum e_Fc_type *Val,
									  IN int LineNum) {
	enum e_ArchToken Token;
	const char* const* PrevArgs = Args;
	Args = ReadToken(Args, &Token, LineNum);
	switch(Token) {
		case AT_FRACTIONAL: *Val = FRACTIONAL; break;
		case AT_ABSOLUTE: *Val = ABSOLUTE; break;
		default: Error(*PrevArgs, LineNum); break;
	}
	return Args;
}


static const char *const * ReadChanWidth(IN const char *const *Args,
						  				 OUT t_chan *Chan,
										 IN int LineNum) {
	enum e_ArchToken Token;
	const char* const* PrevArgs = Args;
	Args = ReadToken(Args, &Token, LineNum);
	switch(Token) {
		case AT_UNIFORM:
			Chan->type = UNIFORM;
			Args = ReadFloat(Args, &Chan->peak, LineNum);
			Chan->width = OPEN;
			Chan->xpeak = OPEN;
			Chan->dc = OPEN;
			break;

		case AT_DELTA:
			Chan->type = DELTA;
			Args = ReadFloat(Args, &Chan->peak, LineNum);
			Chan->width = OPEN;
			Args = ReadFloat(Args, &Chan->xpeak, LineNum);
			Args = ReadFloat(Args, &Chan->dc, LineNum);
			break;

		case AT_PULSE:
		case AT_GAUSSIAN:
			Chan->type = (AT_GAUSSIAN == Token) ? GAUSSIAN : PULSE;
			Args = ReadFloat(Args, &Chan->peak, LineNum);
			Args = ReadFloat(Args, &Chan->width, LineNum);
			Args = ReadFloat(Args, &Chan->xpeak, LineNum);
			Args = ReadFloat(Args, &Chan->dc, LineNum);
			break;

		default:
			Error(*PrevArgs, LineNum);
	}
	return Args;
}


static const char *const * ReadPin(IN const char *const *Args,
								   INOUT struct s_pin **PinList, 
								   IN enum e_pin_type Type,
								   IN int LineNum) {
	const char *const * NextArgs;
	enum e_ArchToken Token;
	struct s_pin *Cur;
	struct s_pin *Tmp;

	Cur = (struct s_pin*) my_malloc(sizeof(struct s_pin));
	Cur->Next = NULL;
	if (*PinList) {
		Tmp = *PinList;
		while (Tmp->Next) Tmp = Tmp->Next;
		Tmp->Next = Cur;
	}
	else {
		*PinList = Cur;
	}

	/* Save the pin type */
	Cur->Type = Type;

	/* Read class */
	Args = ReadTag(Args, AT_CLASS, LineNum);
	Args = ReadInt(Args, &Cur->Class, LineNum);
		
	NextArgs = ReadToken(Args, &Token, LineNum);

	/* Is the pin marked as global */
	if (AT_GLOBAL == Token) {
		Args = NextArgs;
		NextArgs = ReadToken(Args, &Token, LineNum);
		Cur->IsGlobal = TRUE;
	}
	else {
		Cur->IsGlobal = FALSE;
	}

	/* Figure out what side it wants to be on */
	switch (Token) {
		case AT_LEFT: Cur->Side = LEFT; break;
		case AT_TOP: Cur->Side = TOP; break;
		case AT_RIGHT: Cur->Side = RIGHT; break;
		case AT_BOTTOM: Cur->Side = BOTTOM; break;
		default: Error(*Args, LineNum);
	}	

	return NextArgs;
}


static const char *const * ReadInt(IN const char *const *Args, 
								   OUT int *Val,
								   IN int LineNum) {
	if (NULL == *Args) Error(*Args, LineNum);
	if ((**Args > '9') || (**Args < '0')) Error(*Args, LineNum);

	*Val = atoi(*Args);

	return ++Args;
}


static const char *const * ReadFloat(IN const char *const *Args, 
								     OUT float *Val,
								     IN int LineNum) {
	if (NULL == *Args) Error(*Args, LineNum);

	if ( (**Args != '-') && 
		 (**Args != '.') && 
		 ((**Args > '9') || (**Args < '0')) ) {
		Error(*Args, LineNum);
	}
	
	*Val = atof(*Args);

	return ++Args;
}


/* Finds the enum associated with a token. This could be a binary search
 * if you ensure DecodeArchTokenList is sorted but speedup isn't worth
 * hassel of ensuring it is sorted. */
static const char *const * ReadBaseToken(IN const char *const *Args,
										 OUT enum e_ArchBaseToken *Token,
										 IN int LineNum) {
	int i;
		
	/* Empty string is end of tokens marker */
	if (NULL == *Args) Error(*Args, LineNum);

	/* Linear search for the pair */
	for (i = 0; NULL != ArchBaseTokenList[i].Str; ++i) {
		if (strcmp(*Args, ArchBaseTokenList[i].Str) == 0) {
			*Token = ArchBaseTokenList[i].Enum;
			return ++Args;
		}
	}

	*Token = AT_BASE_UNKNOWN;
	return ++Args;
}


static const char *const * ReadToken(IN const char *const *Args,
									 OUT enum e_ArchToken *Token,
									 IN int LineNum) {	
	int i;
		
	/* Empty string is end of tokens marker */
	if (NULL == *Args) Error(*Args, LineNum);

	/* Linear search for the pair */
	for (i = 0; NULL != ArchArgTokenList[i].Str; ++i) {
		if (strcmp(*Args, ArchArgTokenList[i].Str) == 0) {
			*Token = ArchArgTokenList[i].Enum;
			return ++Args;
		}
	}

	*Token = AT_ARG_UNKNOWN;
	return ++Args;
}
